/*$Id: e_elemnt.cc,v 26.43 2007/07/22 07:47:07 al Exp $ -*- C++ -*-
 * Copyright (C) 2001 Albert Davis
 * Author: Albert Davis <aldavis@ieee.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *------------------------------------------------------------------
 * Base class for elements of a circuit
 */
//testing=script 2006.07.12
#include "u_xprobe.h"
#include "e_aux.h"
#include "bm.h"
#include "e_elemnt.h"
/*--------------------------------------------------------------------------*/
ELEMENT::ELEMENT()
  :COMPONENT(),
   _loaditer(0),
   _m0(),
   _m1(),
   _loss0(0.),
   _loss1(0.),
   _acg(0.),
   _y0(),
   _y1(),
   _y2(),
   _ev(0.) 
{
  _n = _nodes;
  assert(_y0.x == 0. && _y0.f0 == 0. && _y0.f1 == 0.);
  assert(_y1 == _y0);
  assert(_y2 == _y0);
}
/*--------------------------------------------------------------------------*/
ELEMENT::ELEMENT(const ELEMENT& p)
  :COMPONENT(p),
   _loaditer(0),
   _m0(),
   _m1(),
   _loss0(p._loss0),
   _loss1(p._loss1),
   _acg(0.),
   _y0(),
   _y1(),
   _y2(),
   _ev(0.)
{
  trace0(long_label().c_str());
  _n = _nodes;
  if (p._n == p._nodes) {
    for (int ii = 0;  ii < NODES_PER_BRANCH;  ++ii) {
      _n[ii] = p._n[ii];
    }
  }else{
    assert(p._nodes);
    // the constructor for a derived class will take care of it
  }
  assert(_y0.x == 0. && _y0.f0 == 0. && _y0.f1 == 0.);
  assert(_y1 == _y0);
  assert(_y2 == _y0);
}
/*--------------------------------------------------------------------------*/
bool ELEMENT::skip_dev_type(CS& cmd)
{
  return cmd.icmatch(dev_type());
}
/*--------------------------------------------------------------------------*/
void ELEMENT::tr_iwant_matrix_passive()
{
  assert(matrix_nodes() == 2);
  assert(is_device());
  //assert(!subckt()); ok for subckt to exist for logic

  assert(_n[OUT1].m_() != INVALID_NODE);
  assert(_n[OUT2].m_() != INVALID_NODE);

  aa.iwant(_n[OUT1].m_(),_n[OUT2].m_());
  lu.iwant(_n[OUT1].m_(),_n[OUT2].m_());
}
/*--------------------------------------------------------------------------*/
void ELEMENT::tr_iwant_matrix_active()
{
  assert(matrix_nodes() == 4);
  assert(is_device());
  assert(!subckt());

  assert(_n[OUT1].m_() != INVALID_NODE);
  assert(_n[OUT2].m_() != INVALID_NODE);
  assert(_n[IN1].m_() != INVALID_NODE);
  assert(_n[IN2].m_() != INVALID_NODE);

  //aa.iwant(_n[OUT1].m_(),_n[OUT2].m_());
  aa.iwant(_n[OUT1].m_(),_n[IN1].m_());
  aa.iwant(_n[OUT1].m_(),_n[IN2].m_());
  aa.iwant(_n[OUT2].m_(),_n[IN1].m_());
  aa.iwant(_n[OUT2].m_(),_n[IN2].m_());
  //aa.iwant(_n[IN1].m_(),_n[IN2].m_());

  //lu.iwant(_n[OUT1].m_(),_n[OUT2].m_());
  lu.iwant(_n[OUT1].m_(),_n[IN1].m_());
  lu.iwant(_n[OUT1].m_(),_n[IN2].m_());
  lu.iwant(_n[OUT2].m_(),_n[IN1].m_());
  lu.iwant(_n[OUT2].m_(),_n[IN2].m_());
  //lu.iwant(_n[IN1].m_(),_n[IN2].m_());
}
/*--------------------------------------------------------------------------*/
void ELEMENT::tr_iwant_matrix_extended()
{
  assert(is_device());
  assert(!subckt());
  assert(ext_nodes() + int_nodes() == matrix_nodes());

  for (int ii = 0;  ii < matrix_nodes();  ++ii) {
    if (_n[ii].m_() >= 0) {
      for (int jj = 0;  jj < ii ;  ++jj) {
	aa.iwant(_n[ii].m_(),_n[jj].m_());
	lu.iwant(_n[ii].m_(),_n[jj].m_());
      }
    }else{itested();
      // node 1 is grounded or invalid
    }
  }
}
/*--------------------------------------------------------------------------*/
void ELEMENT::ac_iwant_matrix_passive()
{
  acx.iwant(_n[OUT1].m_(),_n[OUT2].m_());
}
/*--------------------------------------------------------------------------*/
void ELEMENT::ac_iwant_matrix_active()
{
  //acx.iwant(_n[OUT1].m_(),_n[OUT2].m_());
  acx.iwant(_n[OUT1].m_(),_n[IN1].m_());
  acx.iwant(_n[OUT1].m_(),_n[IN2].m_());
  acx.iwant(_n[OUT2].m_(),_n[IN1].m_());
  acx.iwant(_n[OUT2].m_(),_n[IN2].m_());
  //acx.iwant(_n[IN1].m_(),_n[IN2].m_());
}
/*--------------------------------------------------------------------------*/
void ELEMENT::ac_iwant_matrix_extended()
{
  assert(is_device());
  assert(!subckt());
  assert(ext_nodes() + int_nodes() == matrix_nodes());

  for (int ii = 0;  ii < matrix_nodes();  ++ii) {
    if (_n[ii].m_() >= 0) {
      for (int jj = 0;  jj < ii ;  ++jj) {
	acx.iwant(_n[ii].m_(),_n[jj].m_());
      }
    }else{itested();
      // node 1 is grounded or invalid
    }
  }
}
/*--------------------------------------------------------------------------*/
double ELEMENT::tr_amps()const
{
  return fixzero((_loss0 * tr_outvolts() + _m0.c1 * tr_involts() + _m0.c0),
  		 _m0.c0);
}
/*--------------------------------------------------------------------------*/
double ELEMENT::tr_probe_num(const std::string& x)const
{
  if (pmatch(x, "Vout")) {
    return tr_outvolts();
  }else if (pmatch(x, "VIn")) {
    return tr_involts();
  }else if (pmatch(x, "I")) {
    return tr_amps();
  }else if (pmatch(x, "P")) {
    return tr_amps() * tr_outvolts();
  }else if (pmatch(x, "PD")) {
    double p = tr_amps() * tr_outvolts();
    return (p > 0.) ? p : 0.;
  }else if (pmatch(x, "PS")) {
    double p = tr_amps() * tr_outvolts();
    return (p < 0.) ? -p : 0.;
  }else if (pmatch(x, "INput")) {
    return _y0.x;
  }else if (pmatch(x, "F")) {
    if (_y0.f0 == LINEAR) {
      untested();
      return _y0.x * _y0.f1;
    }else{
      return _y0.f0;
    }
  }else if (pmatch(x, "NV")) {
    return value();
  }else if (pmatch(x, "EV")) {
    return _y0.f1;
  }else if (pmatch(x, "Y")) {
    return _m0.c1;
  }else if (pmatch(x, "EIV")) {
    untested();
    return _m0.x;
  }else if (pmatch(x, "IOFfset")) {
    itested();
    return _m0.c0;
  }else if (pmatch(x, "IPassive")) {
    untested();
    return _m0.c1 * tr_involts();
  }else if (pmatch(x, "ILoss")) {
    untested();
    return _loss0 * tr_outvolts();
//  }else if (pmatch(x, "DIdt")) {
//    double i0  = (_m0.c1  * _m0.x  + _m0.c0);
//    double it1 = (mt1.f1 * mt1.x + mt1.c0);
//    return  (i0 - it1) / (time0 - time1);
//  }else if (pmatch(x, "DTNew")) {
//    return timef - time0;
//  }else if (pmatch(x, "DTOld")) {
//    return time0 - time1;
//  }else if (pmatch(x, "TIMEF")) {
//    return timef;
//  }else if (pmatch(x, "TIME")) {
//    return time0;
//  }else if (pmatch(x, "TIMEO")) {
//    return time1;
  }else if (pmatch(x, "R")) {
    return (_m0.c1!=0.) ? 1/_m0.c1 : MAXDBL;
  }else if (pmatch(x, "Z")) {
    return port_impedance(_n[OUT1], _n[OUT2], lu, (_m0.c1+_loss0));
  }else if (pmatch(x, "ZRAW")) {
    return port_impedance(_n[OUT1], _n[OUT2], lu, 0.);
  }else{
    itested();
    return COMPONENT::tr_probe_num(x);
  }
}
/*--------------------------------------------------------------------------*/
COMPLEX ELEMENT::ac_amps()const
{
  assert(!is_source());
  return (ac_involts() * _acg + ac_outvolts() * _loss0);
}
/*--------------------------------------------------------------------------*/
XPROBE ELEMENT::ac_probe_ext(const std::string& x)const
{
  COMPLEX admittance = (is_source()) ? _loss0 : _acg+_loss0;

  if (pmatch(x, "Vout")) {			/* volts (out) */
    return XPROBE(ac_outvolts());
  }else if (pmatch(x, "VIN")) {			/* volts (in) */
    return XPROBE(ac_involts());
  }else if (pmatch(x, "I")) {			/* amps */
    return XPROBE(ac_amps());
  }else if (pmatch(x, "P")) {			/* complex "power" */
    return XPROBE(ac_outvolts() * conj(ac_amps()), mtREAL, 10.);
  }else if (pmatch(x, "NV")) {			/* nominal value */
    untested();
    return XPROBE(value());
  }else if (pmatch(x, "EV")) {			/* effective value */
    return XPROBE(_ev);
  }else if (pmatch(x, "Y")) {			/* admittance */
    untested();
    return XPROBE(admittance, mtREAL);
  }else if (pmatch(x, "R")) {			/* complex "resistance" */
    if (admittance == 0.) {
      untested();
      return XPROBE(MAXDBL);
    }else{
      untested();
      return XPROBE(1. / admittance);
    }
  }else if (pmatch(x, "Z")) {			/* port impedance */
    return XPROBE(port_impedance(_n[OUT1], _n[OUT2], acx, admittance));
  }else if (pmatch(x, "ZRAW")) {		/* port impedance, raw */
    return XPROBE(port_impedance(_n[OUT1], _n[OUT2], acx, COMPLEX(0.)));
  }else{ 					/* bad parameter */
    return COMPONENT::ac_probe_ext(x);
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
