/*$Id: lang_spice_in.cc,v 26.47 2007/08/18 06:36:19 al Exp $ -*- C++ -*-
 * Copyright (C) 2006 Albert Davis
 * Author: Albert Davis <aldavis@ieee.org>
 *
 * This file is part of "Gnucap", the Gnu Circuit Analysis Package
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
//testing=script 2007.07.13
#include "declare.h"
#include "d_dot.h"
#include "d_coment.h"
#include "d_subckt.h"
#include "e_model.h"
#include "lang_spice.h"
extern DISPATCHER<CARD> device_dispatcher;

// header hack
#include "d_logic.h"
#include "bm.h"
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_comment(CS& cmd, DEV_COMMENT* x)
{
  assert(x);
  x->set(cmd.substr(1));
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_command(CS& cmd, DEV_DOT* x)
{
  assert(x);
  x->set(cmd.fullstring());
  switch (ENV::run_mode) {
  case rPRE_MAIN:	unreachable();	break;
  case rBATCH:		untested();
  case rINTERACTIVE:	untested();
  case rPRESET:
  case rSCRIPT:		
    CARD_LIST* scope = (x->owner()) ? x->owner()->subckt() : &CARD_LIST::card_list;
    cmd.skip1b("."); CMD::cmdproc(cmd, scope); break;
  }
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_paramset(CS& cmd, MODEL_CARD* x)
{
  assert(x);
  cmd.reset();
  cmd >> ".model";
  parse_label(cmd, x);
  parse_type(cmd, x);
  parse_args(cmd, x);
  cmd.check(bWARNING, "what's this?");
}
/*--------------------------------------------------------------------------*/
std::string extract_type(CS& cmd)
{
  std::string s;
  char id_letter = toupper(cmd.peek());
  switch (id_letter) {
  case '\0': {untested();
    s = "";
  } break;
  case '.': {
    cmd.skip();
    cmd >> s;
    notstd::to_lower(&s);
  } break;
  case 'G': {
    int here = cmd.cursor();
    if      (cmd.pscan("VCCAP")){untested(); s = "vccap";}
    else if (cmd.pscan("VCG")) 	{untested(); s = "vcg";}
    else if (cmd.pscan("VCR")) 	{	     s = "vcr";}
    else if (cmd.pscan("VCCS")) {untested(); s = "vccs";}
    else			{	     s = id_letter;}
    cmd.reset(here);
  } break;
  default: {
    s = id_letter;
  } break;
  }
  return s;
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_module(CS& cmd, MODEL_SUBCKT* x)
{
  assert(x);
  cmd.reset();
  (cmd >> ".subckt") || (cmd >> ".macro");
  parse_label(cmd, x);
  parse_ports(cmd, x, x->max_nodes(), x->min_nodes(), 
	      0/*no unnamed par*/, 0/*start*/, true/*all new*/);
  cmd.check(bDANGER, "subckt default parameters not supported yet");

  parse_module_body(cmd, x, x->subckt(), "spice.subckt>");
}
/*--------------------------------------------------------------------------*/
static void skip_pre_stuff(CS& cmd)
{
  cmd.skipbl();
  while (cmd.ematch(CKT_PROMPT)) {untested();
    /* skip any number of copies of the prompt */
  }
  cmd.ematch(ANTI_COMMENT); /* skip mark so spice ignores but gnucap reads */
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_module_body(CS& cmd, MODEL_SUBCKT* x, CARD_LIST* scope,
				   const std::string& prompt)
{
  for (;;) {
    cmd.get_line(prompt);
    
    if (cmd.is_end() || cmd.pmatch(".ENDs") || cmd.pmatch(".EOM")) {
      break;
    }else{
      skip_pre_stuff(cmd);
      int here = cmd.cursor();
      std::string type  = extract_type(cmd);
      if (type == "model") {	
	cmd.reset();
	CMD::cmdproc(cmd, scope);
      }else if (const CARD* proto = device_dispatcher[type]) {
	CARD* new_instance = proto->clone();
	assert(new_instance);
	new_instance->set_owner(x);
	parse_item(cmd, new_instance);
	scope->push_back(new_instance);
      }else{
	DEV_DOT* new_instance = new DEV_DOT;
	new_instance->set_owner(x);
	cmd.reset(here);
	parse_item(cmd, new_instance);
	scope->push_back(new_instance);
      }
    }
  }
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_instance(CS& cmd, COMPONENT* x)
{
  try {
    assert(x);
    cmd.reset().ematch(ANTI_COMMENT);
    
    // ACS dot type specifier
    if (cmd.skip1b('.')) {
      parse_type(cmd, x);
    }else{
    }
    
    parse_label(cmd, x);
    
    if (x->use_obsolete_callback_parse()) {
      parse_element_using_obsolete_callback(cmd, x);
    }else if (DEV_LOGIC* xx = dynamic_cast<DEV_LOGIC*>(x)) {
      parse_logic_using_obsolete_callback(cmd, xx);
    }else{
      parse_ports(cmd, x, x->max_nodes(), x->min_nodes(), x->tail_size(), 0);
      if (x->print_type_in_spice()) {
	parse_type(cmd, x);
      }else{
      }
      parse_args(cmd, x);
    }
  }catch (Exception e) {
    cmd.warn(bDANGER, e.message());
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_element_using_obsolete_callback(CS& cmd, COMPONENT* x)
{
  assert(x);
  ELEMENT* xx = dynamic_cast<ELEMENT*>(x);
  assert(xx);

  parse_ports(cmd, x, x->stop_nodes(), 0,  0,   0);
  //		       max		min tail already_got
  int gotnodes = x->_net_nodes;
  COMMON_COMPONENT* c = NULL;

  if (gotnodes < x->min_nodes()) {
    // HSPICE compatibility kluge.
    // The device type or function type could be stuck between the nodes.
    xx->skip_dev_type(cmd); // (redundant)
    c = EVAL_BM_ACTION_BASE::parse_func_type(cmd);
    parse_ports(cmd, x, x->max_nodes(), x->min_nodes(),
			     0/*no required args*/, gotnodes/*start*/);
  }else{
    // Normal mode.  nodes first, then data.
  }

  if (!c) {
    xx->skip_dev_type(cmd); // (redundant)
    c = new EVAL_BM_COND;
  }else{
  }
  assert(c);

  // we have a blank common of the most general type
  // (or HSPICE kluge)
  // let it continue parsing

  int here = cmd.cursor();
  c->parse_common_obsolete_callback(cmd); //BUG//callback
  if (cmd.stuck(&here)) {untested();
    cmd.warn(bDANGER, "needs a value");
  }else{
  }

  // At this point, there is ALWAYS a common "c", which may have more
  // commons attached to it.  Try to reduce its complexity.
  // "c->deflate()" may return "c" or some simplification of "c".
  
  COMMON_COMPONENT* dc = c->deflate();
  
  // dc == deflated_common
  // It might be just "c".
  // It might be something else that is simpler but equivalent.

  // check for a simple value
  EVAL_BM_VALUE* dvc = dynamic_cast<EVAL_BM_VALUE*>(dc);

  // dvc == deflated value common
  // It might be "dc", if "dc" is a value common.
  // It might be nothing.

  if (dvc && !dvc->has_ext_args()) {
    // If it is a simple value, don't use a common.
    // Just store the value directly.
    x->set_value(dvc->value());
    delete c;
  }else{
    x->attach_common(dc);
    if (dc != c) {
      delete c;
    }else{
    }
  }
  cmd.check(bDANGER, "what's this?");
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_logic_using_obsolete_callback(CS& cmd, COMPONENT* x)
{
  assert(x);

  parse_ports(cmd, x, x->max_nodes(), x->min_nodes(), x->tail_size(), 0/*start*/);

  int incount = x->net_nodes() - x->min_nodes() + 1;
  assert(incount > 0);

  std::string modelname = cmd.ctos(TOKENTERM);

  COMMON_LOGIC* common = 0;
  if      (cmd.pmatch("AND" )) {untested();common = new LOGIC_AND;}
  else if (cmd.pmatch("NAND")) {untested();common = new LOGIC_NAND;}
  else if (cmd.pmatch("OR"  )) {untested();common = new LOGIC_OR;}
  else if (cmd.pmatch("NOR" )) {common = new LOGIC_NOR;}
  else if (cmd.pmatch("XOR" )) {untested();common = new LOGIC_XOR;}
  else if (cmd.pmatch("XNOR")) {untested();common = new LOGIC_XNOR;}
  else if (cmd.pmatch("INV" )) {common = new LOGIC_INV;}
  else {untested();
    cmd.warn(bWARNING,"need and,nand,or,nor,xor,xnor,inv");
    common=new LOGIC_NONE;
  }
  
  assert(common);
  common->incount = incount;
  common->set_modelname(modelname);
  x->attach_common(common);
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_type(CS& cmd, CARD* x)
{
  assert(x);
  std::string new_type;
  cmd >> new_type;
  notstd::to_lower(&new_type);
  x->set_dev_type(new_type);
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_args(CS& cmd, CARD* x)
{
  assert(x);
  if (!x->use_obsolete_callback_parse()) {
    int paren = cmd.skip1b('(');
    int here = cmd.cursor();
    for (int i=0; ; ++i) {
      if (paren && cmd.skip1b(')')) {
	break;
      }else if (!cmd.more()) {
	break;
      }else{
	std::string name, value;
	cmd >> name >> '=' >> value;
	int there = here;
	if (cmd.stuck(&here)) {untested();
	  break;
	}else{
	  try{
	    notstd::to_lower(&name);
	    x->set_param_by_name(name, value);
	  }catch (Exception_No_Match) {
	    COMPONENT* xx = dynamic_cast<COMPONENT*>(x);
	    if (xx && (i == 0) && (xx->param_name(i) == xx->value_name())) {
	      try{
		name = xx->value_name();
		cmd.reset(there);
		cmd >> value;
		x->set_param_by_name(name, value);
	      }catch (Exception_No_Match) {untested();
		cmd.warn(bDANGER, there, "bad arg, ignored, trying to move on");
		break;
	      }
	    }else{itested();
	      cmd.warn(bDANGER, there, "bad parameter, ignored");
	    }
	  }
	}
      }
    }
  }else if (MODEL_CARD* pp = dynamic_cast<MODEL_CARD*>(x)) {
    // used only for "table"
    int paren = cmd.skip1b('(');
    bool in_error = false;
    for (;;) {
      int here = cmd.cursor();
      pp->parse_params_obsolete_callback(cmd);  //BUG//callback//
      if (!cmd.more()) {
	break;
      }else if (paren && cmd.skip1b(')')) {untested();
	break;
      }else if (cmd.stuck(&here)) {untested();
	if (in_error) {untested();
	  // so you don't get two errors on name = value.
	  cmd.skiparg();
	  in_error = false;
	}else{untested();
	  cmd.warn(bDANGER, "bad paramerter -- ignored");
	  cmd.skiparg().skip1b("=");
	  in_error = true;
	}
      }else{
	in_error = false;
      }
    }
  }else{untested();
  }
}
/*--------------------------------------------------------------------------*/
void LANG_SPICE::parse_label(CS& cmd, CARD* x)
{
  assert(x);
  std::string my_name;
  cmd >> my_name;
  x->set_label(my_name);
}
/*--------------------------------------------------------------------------*/
/* parse_ports: parse circuit connections from input string
 * fills in the rest of the array with 0
 * returns the number of nodes actually read
 * The purpose of this complexity is to handle the variants of Spice formats,
 * which might have keys between nodes,
 * an unknown number of nodes with unknown number of args and no delimeters.
 * Consider:  X1 a b c d e f g h i j k
 * Clearly (?) a,b,c,d,e are nodes, f is a subckt name, the rest are args.
 * But how do you know?
 *
 * Args:
 * maxnodes: the maximum number of port nodes to parse.
 *	Stop here, even if it seems there should be more.
 *
 * minnodes: the minimum number of port nodes to parse.
 *	It is an error if there are fewer than this.
 *
 * leave_tail: The number of arguments that are not nodes,
 *	and don't have equal signs, following.
 *	It is an aid to distinguishing the nodes from things that follow.
 *	minnodes wins over leave_tail, but leave_tail wins over maxnodes.
 *	The above example would need leave_tail=6 to parse correctly.
 *	This one: X1 a b c d e f g=h i=j k
 *	where a,b,c,d,e are nodes, f is a subckt or model identifier
 *	would parse correctly with leave_tail=1.
 *	Usual values for leave_tail are 0 or 1.
 *
 * start: start at this number, used for continuing after HSPICE kluge.
 *	Other than that, the only useful value for start is 0.
 *
 * all_new: All node names must be new (not already defined) and unique.
 *	This is used for the header line in .subckt, which starts clean.
 *	The main benefit is to reject duplicates.
 */
void LANG_SPICE::parse_ports(CS& cmd, CARD* x, int maxnodes, int minnodes,
			   int leave_tail, int start, bool all_new)
{
  assert(minnodes <= maxnodes);
  assert(x);

  if (start < maxnodes) {
    int paren = cmd.skip1b('(');
    int num_nodes = 0;
    if (paren) {
      // let the closing paren determine the end of the node list
      num_nodes = maxnodes;
    }else{
      // loop over the tokens to try to guess where the nodes end
      // and other stuff begins
      int here = cmd.cursor();
      int i = start;
      for (;;) {
	cmd.skiparg();
	++i;
	if (cmd.is_end()) {
	  // found the end, no '='
	  if (i <= minnodes) {untested();
	    num_nodes = i;
	  }else if (i <= minnodes + leave_tail) {
	    num_nodes = minnodes;
	  }else if (i <= maxnodes + leave_tail) {
	    num_nodes = i - leave_tail;
	  }else{
	    num_nodes = maxnodes;
	  }
	  break;
	}else if (cmd.skip1b('=')) {
	  // found '=', it's past the end of nodes
	  if (i <= minnodes) {untested();
	    num_nodes = i;
	  }else if (i <= minnodes + leave_tail + 1) {
	    num_nodes = minnodes;
	  }else if (i <= maxnodes + leave_tail + 1) {
	    num_nodes = i - leave_tail - 1;
	  }else{
	    num_nodes = maxnodes;
	  }
	  break;
	}
      }
      cmd.reset(here);
    }
    // got num_nodes

    int ii = start;
    for (;;) {
      int here = cmd.cursor();
      if (paren && cmd.skip1b(')')) {
	--paren;
	break; // done.  have closing paren.
      }else if (ii >= num_nodes) {
	break; // done.  have maxnodes.
      }else if (ii >= maxnodes) {untested();
	unreachable();
	break; // done.  have maxnodes.
      }else if (!cmd.more()) {untested();
	break; // done.  premature end of line.
      }else if (OPT::keys_between_nodes 
		&& cmd.pmatch("POLY")
		|| cmd.pmatch("PWL")
		|| cmd.pmatch("VCCAP")
		|| cmd.pmatch("VCG")
		|| cmd.pmatch("VCR")
		) {
	cmd.reset(here);
	break; // done, found reserved word between nodes
      }else{
	//----------------------
	try{
	  int here = cmd.cursor();
	  std::string node_name;
	  cmd >> node_name;
	  if (cmd.stuck(&here)) {untested();
	    // didn't move, probably a terminator.
	  }else{
	    // legal node name, store it.
	    x->set_port_by_index(ii, node_name);
	  }
	}catch (Exception_Too_Many e) {untested();
	  cmd.warn(bDANGER, here, e.message());
	}
	//----------------------
	if (!(x->node_is_connected(ii))) {untested();
	  break; // illegal node name, might be proper exit.
	}else{
	  if (all_new) {
	    if (x->node_is_grounded(ii)) {untested();
	      cmd.warn(bDANGER, here, "node 0 not allowed here");
	    }else{
	    }
	  }else{
	  }
	  ++ii;
	}
      }
    }
    if (ii < minnodes) {untested();
      cmd.warn(bDANGER, "need " + to_string(minnodes-ii) +" more nodes");
    }else{
    }
    if (paren != 0) {untested();
      cmd.warn(bWARNING, "need )");
    }else{
    }
    //assert(x->_net_nodes == ii);

    // ground unused input nodes
    for (int iii = ii;  iii < minnodes;  ++iii) {untested();
      x->set_port_to_ground(iii);
    }
    //assert(x->_net_nodes >= ii);
  }else{untested();
    assert(x->_net_nodes == 0);
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
